from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Center, Middle
from textual.screen import ModalScreen
from textual.widgets import Label, ProgressBar


class LoadingScreen(ModalScreen[None]):
    """Modal progress screen shown during repo scanning."""

    DEFAULT_CSS = """
    LoadingScreen {
        align: center middle;
    }

    #loading-box {
        width: 60;
        height: 10;
        border: round $accent;
        padding: 1 2;
        background: $surface;
    }

    #loading-label {
        text-align: center;
        width: 100%;
        margin-bottom: 1;
    }

    #loading-progress {
        width: 100%;
    }
    """

    def compose(self) -> ComposeResult:
        with Middle():
            with Center(id="loading-box"):
                yield Label("Scanning repositories…", id="loading-label")
                yield ProgressBar(total=100, show_eta=False, id="loading-progress")

    def update_progress(self, done: int, total: int) -> None:
        bar = self.query_one("#loading-progress", ProgressBar)
        label = self.query_one("#loading-label", Label)
        if total > 0:
            bar.update(total=total, progress=done)
            label.update(f"Scanning repositories… {done}/{total}")
        else:
            label.update("Scanning repositories…")

    def set_phase(self, text: str) -> None:
        self.query_one("#loading-label", Label).update(text)



from __future__ import annotations

from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Center, Middle, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Label, Static

from branchboard.models import BranchInfo


class DetailScreen(ModalScreen[None]):
    """Modal showing full branch/PR details."""

    BINDINGS = [
        ("escape", "dismiss", "Close"),
        ("q", "dismiss", "Close"),
    ]

    DEFAULT_CSS = """
    DetailScreen {
        align: center middle;
    }

    #detail-box {
        width: 80;
        max-height: 80%;
        border: round $accent;
        padding: 1 2;
        background: $surface;
    }

    #detail-title {
        text-style: bold;
        margin-bottom: 1;
        width: 100%;
    }

    #detail-content {
        width: 100%;
    }
    """

    def __init__(self, branch: BranchInfo, **kwargs) -> None:
        super().__init__(**kwargs)
        self._branch = branch

    def compose(self) -> ComposeResult:
        b = self._branch
        with Middle():
            with VerticalScroll(id="detail-box"):
                yield Label(
                    f"{b.repo_name} / {b.branch_name}",
                    id="detail-title",
                )
                yield Static(self._build_content(), id="detail-content")

    def _build_content(self) -> Text:
        b = self._branch
        lines = Text()

        lines.append("State: ", style="bold")
        lines.append(f"{b.state.label}\n", style=b.state.color)

        lines.append("Repo: ", style="bold")
        lines.append(f"{b.repo_path}\n")

        if b.last_commit_date:
            lines.append("Last commit: ", style="bold")
            lines.append(f"{b.last_commit_date.strftime('%Y-%m-%d %H:%M')} — {b.last_commit_msg}\n")

        if b.has_upstream:
            lines.append("Upstream: ", style="bold")
            ahead_behind = []
            if b.commits_ahead:
                ahead_behind.append(f"{b.commits_ahead} ahead")
            if b.commits_behind:
                ahead_behind.append(f"{b.commits_behind} behind")
            lines.append(", ".join(ahead_behind) if ahead_behind else "up to date")
            lines.append("\n")

        if b.is_dirty and b.dirty_files:
            lines.append(f"\nDirty files ({len(b.dirty_files)}):\n", style="bold red")
            for f in b.dirty_files[:30]:
                lines.append(f"  {f}\n", style="red")
            if len(b.dirty_files) > 30:
                lines.append(f"  … and {len(b.dirty_files) - 30} more\n", style="dim")

        if b.pr:
            pr = b.pr
            lines.append(f"\nPR #{pr.number}: ", style="bold")
            lines.append(f"{pr.title}\n")
            lines.append("URL: ", style="bold")
            lines.append(f"{pr.url}\n")
            lines.append("State: ", style="bold")
            lines.append(f"{pr.state}")
            if pr.is_draft:
                lines.append(" (draft)")
            lines.append("\n")
            if pr.review_decision:
                lines.append("Review: ", style="bold")
                lines.append(f"{pr.review_decision}\n")
            if pr.reviewers:
                lines.append("Reviewers: ", style="bold")
                lines.append(", ".join(pr.reviewers))
                lines.append("\n")
            lines.append("Base: ", style="bold")
            lines.append(f"{pr.base_branch}\n")

        lines.append("\n[Esc] close  [o] open PR in browser", style="dim")
        return lines

    def action_dismiss(self) -> None:
        self.dismiss(None)



from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Input, Select
from textual.widget import Widget

from branchboard.models import BranchState


_STATE_OPTIONS: list[tuple[str, str]] = [
    ("All States", "ALL"),
] + [(s.label, s.value) for s in BranchState]


class FilterBar(Widget):
    """Search input + state dropdown."""

    DEFAULT_CSS = """
    FilterBar {
        dock: top;
        height: 3;
        layout: horizontal;
        padding: 0 1;
    }

    FilterBar Input {
        width: 1fr;
        margin-right: 1;
    }

    FilterBar Select {
        width: 28;
    }
    """

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Search repo/branch…", id="search-input")
        yield Select(_STATE_OPTIONS, value="ALL", id="state-select", allow_blank=False)



from __future__ import annotations

from datetime import datetime, timezone
from rich.text import Text
from textual.binding import Binding
from textual.widgets import DataTable

from branchboard.models import BranchInfo, BranchState, STATE_PRIORITY


def _relative_time(dt: datetime | None) -> str:
    if dt is None:
        return ""
    now = datetime.now(timezone.utc)
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    delta = now - dt
    days = delta.days
    if days == 0:
        hours = delta.seconds // 3600
        if hours == 0:
            return f"{delta.seconds // 60}m ago"
        return f"{hours}h ago"
    if days == 1:
        return "yesterday"
    if days < 30:
        return f"{days}d ago"
    months = days // 30
    return f"{months}mo ago"


def _sort_key_for_date(dt: datetime | None) -> float:
    """Return a timestamp for sorting. None -> very old."""
    if dt is None:
        return 0.0
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    return dt.timestamp()


class SortMode:
    PRIORITY = "priority"
    RECENT = "recent"


class BranchTable(DataTable):
    """Color-coded branch table with filtering, sorting, and Vim navigation."""

    DEFAULT_CSS = """
    BranchTable {
        height: 1fr;
    }
    """

    BINDINGS = [
        # Vim single-row movement
        Binding("j", "cursor_down", "Down", show=False),
        Binding("k", "cursor_up", "Up", show=False),
        # Vim jump to top/bottom
        Binding("g", "scroll_top", "Top", show=False),
        Binding("G", "scroll_bottom", "Bottom", show=False),
        # Vim half-page
        Binding("ctrl+d", "half_page_down", "Half Pg Down", show=False),
        Binding("ctrl+u", "half_page_up", "Half Pg Up", show=False),
        # Vim full-page (also standard PageUp/PageDown already handled by DataTable)
        Binding("ctrl+f", "page_down", "Pg Down", show=False),
        Binding("ctrl+b", "page_up", "Pg Up", show=False),
    ]

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._all_branches: list[BranchInfo] = []
        self._filtered: list[BranchInfo] = []
        self._search_text = ""
        self._state_filter = "ALL"
        self._sort_mode = SortMode.PRIORITY

    def on_mount(self) -> None:
        self.cursor_type = "row"
        self.zebra_stripes = True
        self.add_columns("Repo", "Branch", "State", "Last Activity", "PR#", "Details")

    def _visible_row_count(self) -> int:
        """Approximate number of visible rows in the table viewport."""
        # content_size.height gives the viewport height; each row is 1 cell high
        # Subtract 1 for the header row
        return max(1, self.size.height - 1)

    # -- Vim navigation actions --

    def action_scroll_top(self) -> None:
        """Move cursor to the first row (Vim: gg)."""
        if self.row_count > 0:
            self.move_cursor(row=0)

    def action_scroll_bottom(self) -> None:
        """Move cursor to the last row (Vim: G)."""
        if self.row_count > 0:
            self.move_cursor(row=self.row_count - 1)

    def action_half_page_down(self) -> None:
        """Move cursor down by half a page (Vim: Ctrl-d)."""
        if self.row_count == 0:
            return
        half = self._visible_row_count() // 2
        new_row = min(self.cursor_row + half, self.row_count - 1)
        self.move_cursor(row=new_row)

    def action_half_page_up(self) -> None:
        """Move cursor up by half a page (Vim: Ctrl-u)."""
        if self.row_count == 0:
            return
        half = self._visible_row_count() // 2
        new_row = max(self.cursor_row - half, 0)
        self.move_cursor(row=new_row)

    def action_page_down(self) -> None:
        """Move cursor down by a full page (Vim: Ctrl-f)."""
        if self.row_count == 0:
            return
        page = self._visible_row_count()
        new_row = min(self.cursor_row + page, self.row_count - 1)
        self.move_cursor(row=new_row)

    def action_page_up(self) -> None:
        """Move cursor up by a full page (Vim: Ctrl-b)."""
        if self.row_count == 0:
            return
        page = self._visible_row_count()
        new_row = max(self.cursor_row - page, 0)
        self.move_cursor(row=new_row)

    # -- Sort / filter / data --

    @property
    def sort_mode(self) -> str:
        return self._sort_mode

    def toggle_sort(self) -> str:
        """Toggle between priority and recency sort. Returns new mode."""
        if self._sort_mode == SortMode.PRIORITY:
            self._sort_mode = SortMode.RECENT
        else:
            self._sort_mode = SortMode.PRIORITY
        self._apply_filter()
        return self._sort_mode

    def set_branches(self, branches: list[BranchInfo]) -> None:
        self._all_branches = branches
        self._apply_filter()

    def set_search(self, text: str) -> None:
        self._search_text = text.lower()
        self._apply_filter()

    def set_state_filter(self, state: str) -> None:
        self._state_filter = state
        self._apply_filter()

    def _apply_filter(self) -> None:
        filtered = self._all_branches
        if self._search_text:
            filtered = [
                b for b in filtered
                if self._search_text in b.repo_name.lower()
                or self._search_text in b.branch_name.lower()
            ]
        if self._state_filter != "ALL":
            filtered = [b for b in filtered if b.state.value == self._state_filter]

        # Apply sorting
        if self._sort_mode == SortMode.RECENT:
            filtered = sorted(
                filtered,
                key=lambda b: _sort_key_for_date(b.last_commit_date),
                reverse=True,
            )
        else:
            filtered = sorted(
                filtered,
                key=lambda b: (STATE_PRIORITY.get(b.state, 99), b.repo_name, b.branch_name),
            )

        self._filtered = filtered
        self._rebuild_rows()

    def _rebuild_rows(self) -> None:
        self.clear()
        for i, b in enumerate(self._filtered):
            color = b.state.color
            state_text = Text(f" {b.state.label} ", style=f"bold {color}")
            pr_num = Text(f"#{b.pr.number}", style=color) if b.pr else Text("")

            detail = ""
            if b.is_dirty:
                detail = f"{len(b.dirty_files)} dirty files"
            elif b.pr:
                detail = b.pr.title[:50]
            elif b.commits_ahead:
                detail = f"{b.commits_ahead} unpushed commit(s)"

            self.add_row(
                Text(b.repo_name),
                Text(b.branch_name, style=color),
                state_text,
                Text(_relative_time(b.last_commit_date)),
                pr_num,
                Text(detail, style="dim" if b.state == BranchState.STALE else ""),
                key=f"{b.repo_path}:{b.branch_name}:{i}",
            )

    def get_selected_branch(self) -> BranchInfo | None:
        if self.cursor_row is None or self.cursor_row >= len(self._filtered):
            return None
        return self._filtered[self.cursor_row]



from __future__ import annotations

from datetime import datetime, timezone, timedelta
from branchboard.models import BranchInfo, BranchState, STATE_PRIORITY

_STALE_DAYS = 14


def classify_branch(branch: BranchInfo) -> BranchState:
    """Determine the state of a single branch."""
    # Dirty takes highest priority
    if branch.is_dirty:
        return BranchState.DIRTY

    # Unpushed commits (no upstream or ahead of upstream)
    if not branch.has_upstream or (branch.commits_ahead > 0 and branch.pr is None):
        # Check if stale first
        if _is_stale(branch) and branch.pr is None:
            return BranchState.STALE
        if not branch.has_upstream or branch.commits_ahead > 0:
            if branch.pr is None:
                if not branch.has_upstream:
                    return BranchState.UNPUSHED
                return BranchState.UNPUSHED

    # PR-based states
    if branch.pr is not None:
        pr = branch.pr
        if pr.state == "MERGED":
            return BranchState.PR_MERGED
        if pr.state == "CLOSED":
            # Closed but not merged — treat as stale
            return BranchState.STALE
        # OPEN PR
        if pr.is_draft:
            return BranchState.PR_DRAFT
        if pr.review_decision == "APPROVED":
            return BranchState.PR_APPROVED
        if pr.review_decision == "CHANGES_REQUESTED":
            return BranchState.PR_CHANGES_REQ
        return BranchState.PR_OPEN

    # Has upstream but no PR
    if branch.has_upstream:
        if _is_stale(branch):
            return BranchState.STALE
        return BranchState.NO_PR

    # Fallback
    if _is_stale(branch):
        return BranchState.STALE
    return BranchState.NO_PR


def _is_stale(branch: BranchInfo) -> bool:
    if branch.last_commit_date is None:
        return True
    cutoff = datetime.now(timezone.utc) - timedelta(days=_STALE_DAYS)
    commit_date = branch.last_commit_date
    if commit_date.tzinfo is None:
        commit_date = commit_date.replace(tzinfo=timezone.utc)
    return commit_date < cutoff


def classify_all(branches: list[BranchInfo]) -> None:
    """Classify all branches in place and sort by priority."""
    for b in branches:
        b.state = classify_branch(b)
    branches.sort(key=lambda b: (STATE_PRIORITY.get(b.state, 99), b.repo_name, b.branch_name))



from __future__ import annotations

import asyncio
import os
from datetime import datetime, timezone
from pathlib import Path

from branchboard.models import BranchInfo

# Scan up to 10 repos concurrently
_GIT_SEMAPHORE = asyncio.Semaphore(10)


async def _run(cmd: list[str], cwd: str) -> tuple[int, str]:
    proc = await asyncio.create_subprocess_exec(
        *cmd,
        cwd=cwd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, _ = await proc.communicate()
    return proc.returncode, stdout.decode(errors="replace").strip()


def discover_repos(root: str, max_depth: int = 3) -> list[str]:
    """Walk *root* up to *max_depth* levels looking for .git directories."""
    repos: list[str] = []
    root_path = Path(root).resolve()
    for dirpath, dirnames, _ in os.walk(root_path):
        depth = len(Path(dirpath).relative_to(root_path).parts)
        if depth >= max_depth:
            dirnames.clear()
            continue
        if ".git" in dirnames:
            repos.append(dirpath)
            dirnames.clear()  # don't descend into a repo
    repos.sort()
    return repos


async def _get_current_branch(repo: str) -> tuple[str, bool]:
    """Return (branch_name, is_detached)."""
    rc, out = await _run(["git", "symbolic-ref", "--short", "HEAD"], repo)
    if rc == 0:
        return out, False
    # Detached HEAD — get short sha
    _, out = await _run(["git", "rev-parse", "--short", "HEAD"], repo)
    return out, True


async def _get_dirty_files(repo: str) -> list[str]:
    _, out = await _run(["git", "status", "--porcelain"], repo)
    return [line for line in out.splitlines() if line.strip()] if out else []


async def _get_branches(repo: str) -> list[dict]:
    """Return list of branch dicts from for-each-ref."""
    fmt = "%(refname:short)%09%(upstream:short)%09%(upstream:track)%09%(committerdate:iso-strict)%09%(subject)"
    _, out = await _run(
        ["git", "for-each-ref", "--format", fmt, "refs/heads/"],
        repo,
    )
    branches = []
    for line in out.splitlines():
        if not line.strip():
            continue
        parts = line.split("\t")
        if len(parts) < 5:
            parts += [""] * (5 - len(parts))
        name, upstream, track, date_str, subject = parts[:5]
        # Skip main/master/develop unless they have dirty changes (handled separately)
        ahead = behind = 0
        if track:
            import re
            m = re.search(r"ahead (\d+)", track)
            if m:
                ahead = int(m.group(1))
            m = re.search(r"behind (\d+)", track)
            if m:
                behind = int(m.group(1))

        commit_date = None
        if date_str:
            try:
                commit_date = datetime.fromisoformat(date_str)
            except ValueError:
                pass

        branches.append({
            "name": name,
            "upstream": upstream,
            "ahead": ahead,
            "behind": behind,
            "commit_date": commit_date,
            "subject": subject,
        })
    return branches


async def _get_remote_url(repo: str) -> str:
    _, out = await _run(["git", "remote", "get-url", "origin"], repo)
    return out


async def scan_repo(repo_path: str) -> list[BranchInfo]:
    """Scan a single repo for branch information."""
    async with _GIT_SEMAPHORE:
        repo_name = os.path.basename(repo_path)

        current_branch, is_detached = await _get_current_branch(repo_path)
        dirty_files = await _get_dirty_files(repo_path)
        branches_raw = await _get_branches(repo_path)
        remote_url = await _get_remote_url(repo_path)

        results: list[BranchInfo] = []
        for br in branches_raw:
            name = br["name"]
            # Skip default branches unless they're the current dirty branch
            if name in ("main", "master", "develop"):
                if not (name == current_branch and dirty_files):
                    continue

            is_current = name == current_branch
            info = BranchInfo(
                repo_name=repo_name,
                repo_path=repo_path,
                branch_name=name,
                is_current=is_current,
                is_dirty=is_current and bool(dirty_files),
                dirty_files=dirty_files if is_current else [],
                has_upstream=bool(br["upstream"]),
                commits_ahead=br["ahead"],
                commits_behind=br["behind"],
                last_commit_date=br["commit_date"],
                last_commit_msg=br["subject"],
                remote_url=remote_url,
            )
            results.append(info)

        # If detached HEAD with dirty changes, add an entry
        if is_detached and dirty_files:
            results.append(BranchInfo(
                repo_name=repo_name,
                repo_path=repo_path,
                branch_name=f"(detached:{current_branch})",
                is_current=True,
                is_dirty=True,
                dirty_files=dirty_files,
                remote_url=remote_url,
            ))

        return results


async def scan_all_repos(
    root: str,
    on_progress: asyncio.Future | None = None,
) -> list[BranchInfo]:
    """Discover repos under *root* and scan them all concurrently.

    *on_progress* — if provided, should be an async callback(done, total).
    """
    repos = discover_repos(root)
    total = len(repos)
    done = 0
    all_branches: list[BranchInfo] = []

    async def _scan_one(repo: str) -> list[BranchInfo]:
        nonlocal done
        result = await scan_repo(repo)
        done += 1
        if on_progress:
            await on_progress(done, total)
        return result

    tasks = [_scan_one(r) for r in repos]
    results = await asyncio.gather(*tasks)
    for branch_list in results:
        all_branches.extend(branch_list)
    return all_branches



from __future__ import annotations

import enum
from dataclasses import dataclass, field
from datetime import datetime


class BranchState(enum.Enum):
    DIRTY = "DIRTY"
    UNPUSHED = "UNPUSHED"
    PR_CHANGES_REQ = "PR_CHANGES_REQ"
    PR_OPEN = "PR_OPEN"
    PR_APPROVED = "PR_APPROVED"
    PR_DRAFT = "PR_DRAFT"
    NO_PR = "NO_PR"
    STALE = "STALE"
    PR_MERGED = "PR_MERGED"

    @property
    def color(self) -> str:
        return _STATE_COLORS[self]

    @property
    def label(self) -> str:
        return _STATE_LABELS[self]


# Use explicit hex colors for better terminal contrast
_STATE_COLORS: dict[BranchState, str] = {
    BranchState.DIRTY: "#ff5555",          # bright red
    BranchState.UNPUSHED: "#ff7979",       # lighter red/salmon
    BranchState.PR_CHANGES_REQ: "#ffb86c", # orange
    BranchState.PR_OPEN: "#f1fa8c",        # yellow
    BranchState.PR_APPROVED: "#50fa7b",    # green
    BranchState.PR_DRAFT: "#8be9fd",       # cyan
    BranchState.NO_PR: "#bd93f9",          # purple
    BranchState.STALE: "#6272a4",          # muted grey-blue
    BranchState.PR_MERGED: "#69ff94",      # bright green
}

_STATE_LABELS: dict[BranchState, str] = {
    BranchState.DIRTY: "Dirty",
    BranchState.UNPUSHED: "Unpushed",
    BranchState.PR_CHANGES_REQ: "Changes Req",
    BranchState.PR_OPEN: "PR Open",
    BranchState.PR_APPROVED: "PR Approved",
    BranchState.PR_DRAFT: "PR Draft",
    BranchState.NO_PR: "No PR",
    BranchState.STALE: "Stale",
    BranchState.PR_MERGED: "PR Merged",
}

# Priority for sorting — lower = more important (shows first)
STATE_PRIORITY: dict[BranchState, int] = {
    BranchState.DIRTY: 0,
    BranchState.UNPUSHED: 1,
    BranchState.PR_CHANGES_REQ: 2,
    BranchState.PR_OPEN: 3,
    BranchState.PR_APPROVED: 4,
    BranchState.PR_DRAFT: 5,
    BranchState.NO_PR: 6,
    BranchState.STALE: 7,
    BranchState.PR_MERGED: 8,
}

# State label colors for the summary bar (bolder/more visible)
STATE_BADGE_STYLES: dict[BranchState, str] = {
    BranchState.DIRTY: "bold #ff5555",
    BranchState.UNPUSHED: "bold #ff7979",
    BranchState.PR_CHANGES_REQ: "bold #ffb86c",
    BranchState.PR_OPEN: "bold #f1fa8c",
    BranchState.PR_APPROVED: "bold #50fa7b",
    BranchState.PR_DRAFT: "bold #8be9fd",
    BranchState.NO_PR: "bold #bd93f9",
    BranchState.STALE: "#6272a4",
    BranchState.PR_MERGED: "bold #69ff94",
}


@dataclass
class PRInfo:
    number: int
    title: str
    state: str  # OPEN, CLOSED, MERGED
    url: str
    is_draft: bool = False
    review_decision: str = ""  # APPROVED, CHANGES_REQUESTED, REVIEW_REQUIRED, ""
    reviewers: list[str] = field(default_factory=list)
    created_at: str = ""
    updated_at: str = ""
    head_branch: str = ""
    base_branch: str = ""


@dataclass
class BranchInfo:
    repo_name: str
    repo_path: str
    branch_name: str
    is_current: bool = False
    is_dirty: bool = False
    dirty_files: list[str] = field(default_factory=list)
    has_upstream: bool = False
    commits_ahead: int = 0
    commits_behind: int = 0
    last_commit_date: datetime | None = None
    last_commit_msg: str = ""
    remote_url: str = ""
    pr: PRInfo | None = None
    state: BranchState = BranchState.NO_PR



from __future__ import annotations

import json
import os
import time
from pathlib import Path

_CACHE_DIR = Path.home() / ".cache" / "branchboard"
_TTL_SECONDS = 5 * 60  # 5 minutes


def _cache_path(key: str) -> Path:
    return _CACHE_DIR / f"{key}.json"


def load_cache(key: str) -> object | None:
    """Load cached data if it exists and is fresh."""
    path = _cache_path(key)
    if not path.exists():
        return None
    try:
        mtime = path.stat().st_mtime
        if time.time() - mtime > _TTL_SECONDS:
            return None
        with open(path) as f:
            return json.load(f)
    except (OSError, json.JSONDecodeError):
        return None


def save_cache(key: str, data: object) -> None:
    """Save data to cache."""
    _CACHE_DIR.mkdir(parents=True, exist_ok=True)
    path = _cache_path(key)
    try:
        with open(path, "w") as f:
            json.dump(data, f)
    except OSError:
        pass


def clear_cache() -> None:
    """Remove all cached files."""
    if _CACHE_DIR.exists():
        for f in _CACHE_DIR.iterdir():
            if f.suffix == ".json":
                f.unlink(missing_ok=True)



from __future__ import annotations

from rich.text import Text
from textual.widgets import Static

from branchboard.models import BranchInfo, BranchState, STATE_BADGE_STYLES


class SummaryBar(Static):
    """Header bar showing counts by state."""

    DEFAULT_CSS = """
    SummaryBar {
        dock: top;
        height: 1;
        background: #282a36;
        padding: 0 1;
    }
    """

    def update_counts(self, branches: list[BranchInfo]) -> None:
        counts: dict[BranchState, int] = {}
        for b in branches:
            counts[b.state] = counts.get(b.state, 0) + 1

        parts = Text()
        parts.append(f" {len(branches)} branches ", style="bold white")
        parts.append(" ")

        show_order = [
            BranchState.DIRTY,
            BranchState.UNPUSHED,
            BranchState.PR_CHANGES_REQ,
            BranchState.PR_OPEN,
            BranchState.PR_APPROVED,
            BranchState.PR_DRAFT,
            BranchState.NO_PR,
            BranchState.STALE,
            BranchState.PR_MERGED,
        ]
        for state in show_order:
            count = counts.get(state, 0)
            if count > 0:
                parts.append(f" {count} {state.label} ", style=STATE_BADGE_STYLES[state])
                parts.append(" ")

        self.update(parts)



from __future__ import annotations

import asyncio
import webbrowser
from pathlib import Path

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.widgets import DataTable, Footer, Input, Select

from branchboard.cache import clear_cache
from branchboard.classify import classify_all
from branchboard.github import fetch_all_prs
from branchboard.models import BranchInfo
from branchboard.scanner import scan_all_repos
from branchboard.screens.detail import DetailScreen
from branchboard.screens.loading import LoadingScreen
from branchboard.widgets.branch_table import BranchTable
from branchboard.widgets.filter_bar import FilterBar
from branchboard.widgets.summary_bar import SummaryBar


class GitFleetApp(App):
    """Git Branch Dashboard TUI."""

    TITLE = "branchboard"
    CSS_PATH = "app.tcss"

    BINDINGS = [
        Binding("q", "quit", "Quit"),
        Binding("r", "refresh", "Refresh", priority=True),
        Binding("o", "open_pr", "Open PR", priority=True),
        Binding("s", "toggle_sort", "Sort", priority=True),
        Binding("slash", "focus_search", "/ Search", show=False),
        Binding("escape", "focus_table", "Esc Table", show=False),
    ]

    def __init__(self, scan_path: str, use_cache: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self._scan_path = scan_path
        self._use_cache = use_cache
        self._branches: list[BranchInfo] = []

    def compose(self) -> ComposeResult:
        yield SummaryBar(id="summary")
        yield FilterBar()
        with Vertical(id="main-container"):
            yield BranchTable(id="branch-table")
        yield Footer()

    async def on_mount(self) -> None:
        await self._do_scan()

    async def _do_scan(self) -> None:
        loading = LoadingScreen()
        self.push_screen(loading)

        try:
            # Phase 1: git scanning
            async def on_git_progress(done: int, total: int) -> None:
                loading.update_progress(done, total)

            branches = await scan_all_repos(self._scan_path, on_progress=on_git_progress)

            # Phase 2: GitHub PR fetching
            loading.set_phase("Fetching PR data from GitHub…")
            await fetch_all_prs(branches, use_cache=self._use_cache)

            # Phase 3: classify
            classify_all(branches)
            self._branches = branches
        finally:
            self.pop_screen()

        # Update UI
        self._update_display()
        # Focus the table so keyboard shortcuts work
        self.query_one("#branch-table", BranchTable).focus()

    def _update_display(self) -> None:
        table = self.query_one("#branch-table", BranchTable)
        table.set_branches(self._branches)
        summary = self.query_one("#summary", SummaryBar)
        summary.update_counts(self._branches)

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        self.action_show_detail()

    def on_input_changed(self, event: Input.Changed) -> None:
        if event.input.id == "search-input":
            table = self.query_one("#branch-table", BranchTable)
            table.set_search(event.value)

    def on_select_changed(self, event: Select.Changed) -> None:
        if event.select.id == "state-select":
            table = self.query_one("#branch-table", BranchTable)
            table.set_state_filter(str(event.value))

    async def action_refresh(self) -> None:
        clear_cache()
        self._use_cache = False
        await self._do_scan()

    def action_open_pr(self) -> None:
        table = self.query_one("#branch-table", BranchTable)
        branch = table.get_selected_branch()
        if branch and branch.pr and branch.pr.url:
            webbrowser.open(branch.pr.url)
        else:
            self.notify("No PR URL for this branch", severity="warning")

    def action_show_detail(self) -> None:
        table = self.query_one("#branch-table", BranchTable)
        branch = table.get_selected_branch()
        if branch:
            self.push_screen(DetailScreen(branch))

    def action_toggle_sort(self) -> None:
        table = self.query_one("#branch-table", BranchTable)
        new_mode = table.toggle_sort()
        label = "Most Recent" if new_mode == "recent" else "Priority"
        self.notify(f"Sort: {label}", timeout=2)

    def action_focus_search(self) -> None:
        self.query_one("#search-input", Input).focus()

    def action_focus_table(self) -> None:
        self.query_one("#branch-table", BranchTable).focus()

    def _set_state_filter(self, state: str) -> None:
        select = self.query_one("#state-select", Select)
        select.value = state
        table = self.query_one("#branch-table", BranchTable)
        table.set_state_filter(state)



from __future__ import annotations

import argparse
import os
import shutil
import sys


def _check_dependencies() -> None:
    """Verify external CLI tools are available before starting."""
    if not shutil.which("git"):
        print("error: git is not installed or not on PATH", file=sys.stderr)
        print("  Install: https://git-scm.com/downloads", file=sys.stderr)
        sys.exit(1)

    if not shutil.which("gh"):
        print("error: GitHub CLI (gh) is not installed or not on PATH", file=sys.stderr)
        print()
        print("  branchboard uses `gh` to fetch PR data from GitHub.", file=sys.stderr)
        print()
        print("  Install:", file=sys.stderr)
        print("    macOS:  brew install gh", file=sys.stderr)
        print("    Linux:  https://github.com/cli/cli/blob/trunk/docs/install_linux.md", file=sys.stderr)
        print("    Other:  https://cli.github.com/", file=sys.stderr)
        print()
        print("  Then authenticate:  gh auth login", file=sys.stderr)
        sys.exit(1)


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="branchboard",
        description="Git Branch Dashboard TUI — scan repos and show branch/PR status",
    )
    parser.add_argument(
        "--path",
        default=os.path.expanduser("~/Work/repos"),
        help="Root directory to scan for git repos (default: ~/Work/repos)",
    )
    parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Bypass the JSON file cache and fetch fresh data",
    )
    args = parser.parse_args()

    _check_dependencies()

    from branchboard.app import GitFleetApp

    app = GitFleetApp(scan_path=args.path, use_cache=not args.no_cache)
    app.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import json
import re
from branchboard.models import BranchInfo, PRInfo
from branchboard.cache import load_cache, save_cache

_GH_SEMAPHORE = asyncio.Semaphore(5)

_GH_FIELDS = [
    "number", "title", "state", "url", "isDraft",
    "reviewDecision", "reviewRequests", "createdAt", "updatedAt",
    "headRefName", "baseRefName",
]


def _extract_owner_repo(remote_url: str) -> str | None:
    """Extract 'owner/repo' from a git remote URL."""
    # SSH: git@github.com:owner/repo.git
    m = re.match(r"git@github\.com:(.+?)(?:\.git)?$", remote_url)
    if m:
        return m.group(1)
    # HTTPS: https://github.com/owner/repo.git
    m = re.match(r"https://github\.com/(.+?)(?:\.git)?$", remote_url)
    if m:
        return m.group(1)
    return None


async def _run_gh(args: list[str]) -> tuple[int, str]:
    proc = await asyncio.create_subprocess_exec(
        "gh", *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, _ = await proc.communicate()
    return proc.returncode, stdout.decode(errors="replace").strip()


async def _fetch_prs_for_repo(owner_repo: str, use_cache: bool) -> list[dict]:
    """Fetch PRs for a single GitHub repo."""
    if use_cache:
        cached = load_cache(f"prs_{owner_repo.replace('/', '_')}")
        if cached is not None:
            return cached

    async with _GH_SEMAPHORE:
        fields = ",".join(_GH_FIELDS)
        rc, out = await _run_gh([
            "pr", "list",
            "--repo", owner_repo,
            "--state", "all",
            "--author", "@me",
            "--limit", "100",
            "--json", fields,
        ])
        if rc != 0 or not out:
            return []
        try:
            prs = json.loads(out)
        except json.JSONDecodeError:
            return []

    save_cache(f"prs_{owner_repo.replace('/', '_')}", prs)
    return prs


def _match_pr(branch_name: str, prs: list[dict]) -> dict | None:
    """Find the best matching PR for a branch."""
    # Prefer open PRs, then most recent
    matches = [p for p in prs if p.get("headRefName") == branch_name]
    if not matches:
        return None
    # Sort: open first, then by updatedAt descending
    def sort_key(p: dict) -> tuple:
        is_open = 0 if p.get("state") == "OPEN" else 1
        return (is_open, p.get("updatedAt", ""))
    matches.sort(key=sort_key)
    return matches[0]


def _pr_dict_to_info(pr: dict) -> PRInfo:
    reviewers = []
    for req in pr.get("reviewRequests", []):
        name = req.get("login") or req.get("name", "")
        if name:
            reviewers.append(name)
    return PRInfo(
        number=pr.get("number", 0),
        title=pr.get("title", ""),
        state=pr.get("state", ""),
        url=pr.get("url", ""),
        is_draft=pr.get("isDraft", False),
        review_decision=pr.get("reviewDecision") or "",
        reviewers=reviewers,
        created_at=pr.get("createdAt", ""),
        updated_at=pr.get("updatedAt", ""),
        head_branch=pr.get("headRefName", ""),
        base_branch=pr.get("baseRefName", ""),
    )


async def fetch_all_prs(
    branches: list[BranchInfo],
    use_cache: bool = True,
) -> None:
    """Fetch PR data for all branches, mutating branch.pr in place."""
    # Collect unique owner/repo from remote URLs
    remote_map: dict[str, str] = {}  # owner_repo -> remote_url
    for b in branches:
        if b.remote_url:
            owner_repo = _extract_owner_repo(b.remote_url)
            if owner_repo:
                remote_map[owner_repo] = b.remote_url

    # Fetch PRs for each unique repo
    pr_data: dict[str, list[dict]] = {}
    tasks = []
    for owner_repo in remote_map:
        async def _fetch(or_=owner_repo):
            pr_data[or_] = await _fetch_prs_for_repo(or_, use_cache)
        tasks.append(_fetch())

    await asyncio.gather(*tasks)

    # Match PRs to branches
    for b in branches:
        if not b.remote_url:
            continue
        owner_repo = _extract_owner_repo(b.remote_url)
        if not owner_repo or owner_repo not in pr_data:
            continue
        pr = _match_pr(b.branch_name, pr_data[owner_repo])
        if pr:
            b.pr = _pr_dict_to_info(pr)
